/*
 * Module Name: blg_interface.c
 *
 * Description: Interface to BLG code from Hager and Zhang.
 *
 * Copyright: Paulo J. S. Silva <pjssilva@gmail.com> 2013.
 */

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "blg_interface.h"
#include "../../../third_party/BLG-1.2/BLG.h"

/*
 * Function Name: blg_initial_multiplier
 *
 * Description: Estimate an initial multiplier from an approximate
 *     solution.
 *
 * Input:
 *     cqk_problem *restrict p: description of the cont. quad. knapsack.
 *     double *restrict x: approximate solution.
 *
 * Output:
 *     double *restrict y: at output a copy of the projected point (p->a).
 *
 * Return value: the desired estimate.
 *
 * Obs: The multiplier is computed using the approximate to estimate
 * the active face obtained from the given solution. It returns the
 * mulplier associated to the problem of projecting into the
 * hyperplane within this face as suggested by the variable fixing
 * methods.
 */
double blg_initial_multiplier(cqk_problem *restrict p, double *restrict x,
                              double *restrict y) {

    double r = p->r;
    double sum_slopes = 0.0;
    double sum_cts = 0.0;
    /* Estimate the initial multiplier from a given solution. */
    if (x != NULL) {
        for (unsigned i = 0; i < p->n; ++i) {
            if (x[i] == p->low[i] || x[i] == p->up[i])
                r -= x[i];
            else {
                sum_cts += p->a[i];
                sum_slopes += 1.0;
            }
            y[i] = p->a[i];
        }
        /* If all the variables were biding, a very rare and
           unexpected situation, throw the computed values away and
           try again now ignoring all the bounds. */
        if (sum_slopes == 0.0) {
            r = p->r;
            for(unsigned i = 0; i < p->n; ++i) {
                sum_cts += p->a[i];
            }
            sum_slopes = p->n;
        }
    }
    else {
        for(unsigned i = 0; i < p->n; ++i) {
            sum_cts += p->a[i];
            y[i] = p->a[i];
        }
        sum_slopes = p->n;
    }

    return -(r - sum_cts)/(sum_slopes);
}

/*
 * Function Name: blg_interface
 *
 * Description: Interface to BLGnapsack method from Hager and
 *     Zhang. You can get it at:
 *
 *     http://www.math.ufl.edu/~hager/papers/CG/Archive/
 *
 * Input:
 *     cqk_problem *p: the problem description
 *     double *x0: If non NULL, try to use it as starting point (it is
 *         a primal method).
 *
 * Output:
 *     double *restrict x: solution vector.
 *
 * WARNING: BLGnapsack only work for Euclidean projection, that is
 *    p->d must be composed only of ones. Do not use it otherwise.
 *
 * Return value: 0 if solution found or -1 in case of failure.
 */
int blg_interface(cqk_problem *restrict p, double *x0, double *x) {

    /* Allocate working area */
    BLGcom Com;
    Com.w1 = (BLGFLOAT *) malloc(2*p->n*sizeof(BLGFLOAT));
    Com.iw1 = (int *) malloc(2*p->n*sizeof(int));
    Com.iw2 = (int *) malloc(2*p->n*sizeof(int));
    BLGparm  Parm;
    BLGdefault(&Parm);
    Parm.a_is_1 = BLGTRUE;
    Com.Parm = &Parm;

    if (!Com.w1 || !Com.iw1 || !Com.iw2) {
        fprintf(stderr, "Memory allocation error, line %d, file %s\n",
                __LINE__, __FILE__);
        exit(1);
    }

    /* Initialization */
    Com.anorm = p->n;
    double lambda = blg_initial_multiplier(p, x0, x);

    /* Call BLGnapsack */
    int result = BLGnapsack(x, &lambda, NULL, p->r, p->low, p->up, p->n, &Com);

    free(Com.w1);
    free(Com.iw1);
    free(Com.iw2);

    return result;
}

