'''
This is a translation of the readMNIST.m matlab script from Siddharth Hegde to python. 

The original script can be found in

http://www.mathworks.com/matlabcentral/fileexchange/27675-read-digits-and-labels-from-mnist-database/content/readMNIST.m 

%
% Description:
% Read digits and labels from raw MNIST data files
% File format as specified on http://yann.lecun.com/exdb/mnist/
% Note: The 4 pixel padding around the digits will be removed
%       Pixel values will be normalised to the [0...1] range
%
% Usage:
% [imgs labels] = readMNIST(imgFile, labelFile, readDigits, offset)
%
% Parameters:
% imgFile = name of the image file
% labelFile = name of the label file
% readDigits = number of digits to be read
% offset = skips the first offset number of digits before reading starts
%
% Returns:
% imgs = 20 x 20 x readDigits sized matrix of digits
% labels = readDigits x 1 matrix containing labels for each digit
%
'''
import scipy as sp
import sys

def readint32(handle):
    '''Read a int32 number in binary format from file handle.
    
    For some reason the int32 bytes are reversed from the order
    expected by Python, so we read 4 bytes and reverse it.
    '''
    s = handle.read(4) 
    a = sp.fromstring(s[::-1], 'i') 
    return a[0]

def readMNIST(imgFile, labelFile, readDigits, offset):
    
    # Read digits
    fid = file(imgFile, 'rb')
    header = readint32(fid)
    if header != 2051:
        print >>sys.stderr, 'Invalid image file header'
        sys.exit(1)
    
    count = readint32(fid)
    if count < readDigits+offset:
        print >>sys.stderr, 'Trying to read too many digits'
        sys.exit(1)
    
    h = readint32(fid)
    w = readint32(fid)
    
    if offset > 0:
        fid.seek(w*h*offset, 1)
    
    imgs = sp.empty((h, w, readDigits), dtype='uint8')
    
    for i in range(readDigits):
        for y in range(h):
            imgs[y,:,i] = sp.fromfile(fid, 'uint8', w)
    fid.close()

    # Read digit labels
    fid = file(labelFile, 'rb')
    header = readint32(fid)
    if header != 2049:
        print >>sys.stderr, 'Invalid label file header'
        sys.exit(1)

    count = readint32(fid)
    if count < readDigits + offset:
        print >>sys.stderr, 'Trying to read too many digits'
    
    if offset > 0:
        fid.seek(offset, 1)
    
    labels = sp.fromfile(fid, 'uint8', readDigits)
    fid.close()
    
    # Calc avg digit and count
    imgs = trimDigits(imgs, 4)
    imgs = normalizePixValue(imgs)

    return imgs, labels

def trimDigits(digitsIn, border):
    dSize = digitsIn.shape
    digits = sp.empty((dSize[0]-(border*2), dSize[1]-(border*2), dSize[2]))
    for i in range(dSize[2]):
        digits[:,:,i] = digitsIn[border:dSize[0]-border, border:dSize[1]-border, i]
    return digits

def normalizePixValue(digits):
    return digits/255.0
