'''
Summarize the test results in a LaTeX table.
'''

import sys
from platform import architecture

LATEXSTART=r'''
\documentclass{article}

% Some mathematical macros.
\usepackage{amsfonts}
\usepackage{amssymb,amsopn,amsgen,amsbsy,amsmath}

\begin{document}
'''
LATEXEND=r'''
\end{document}
'''

HEADTABLE1 = r'''\begin{table}
\begin{center}
  \begin{scriptsize}
    \begin{tabular}{|l|c||rr|rr|rr|rr|rr|rr|}
      \hline
      {\bf Train set} & $n$ & \multicolumn{2}{c|}{{\bf Newton}} & 
      \multicolumn{2}{c|}{{\bf Newton NC}} & 
      \multicolumn{2}{c|}{{\bf Secant}} & 
      \multicolumn{2}{c|}{{\bf BLG }} &
      \multicolumn{2}{c|}{{\bf Var. Fixing}} & 
      \multicolumn{2}{c|}{{\bf Median Search}} \\ \hline
      & & Iter. & Time & Iter. & Time & Iter. & Time & Iter. & Time & Iter. & Time & Iter. & Time \\ \hline
'''

BUTTONTABLE1 = r'''    \end{tabular}
  \end{scriptsize}
  \caption{Results for the projection in SVM training. The values
    reported are the average number of iterations and average computing
    time in milliseconds among all the projections performed while
    solving~\eqref{eq:svm} using the SPG code.}
  \label{tab:svm}
\end{center}
\end{table}
'''

HEADTABLE2 = r'''\begin{table}
\begin{center}
  \begin{small}
    \begin{tabular}{|l|c||rr|rr|rr|rr|}
      \hline
      {\bf Training set} & $n$ & \multicolumn{2}{c|}{{\bf Newton}} & 
      \multicolumn{2}{c|}{{\bf Newton NC}} &
      \multicolumn{2}{c|}{{\bf Secant}} &
      \multicolumn{2}{c|}{{\bf BLG}} \\ \hline
      & & Iter. & Time & Iter. & Time & Iter. & Time & Iter. & Time \\ \hline
'''

BUTTONTABLE2 = r'''    \end{tabular}
  \end{small}
  \caption{Results for the projections in SVM traning using hot
    starts. The computing time required to estimate the initial
    active face and the corresponding multiplier is included in the
    reported measurements.}
  \label{tab:hotstart}
\end{center}
\end{table}
'''

def get_last(input):
    line = input.readline()
    return line.strip().split()[-1]

def read_projection_result(input):
    '''Read the result of a single projection test.

    Each test should look like

    -------------------- Projection newton
    Reading problem data...
    Calling SPG...
    Number of projections: 864
    Mean steps: 5.047454
    Maximum steps: 17
    Mean time: 3.990741e-05
    SPG Statistics: 1065 431 439 -4.247959e+02 9.207447e-06 0
    '''

    proj_name = get_last(input)
    input.readline()
    input.readline()
    input.readline()
    mean_steps = float(get_last(input))
    input.readline()
    mean_time = float(get_last(input))
    input.readline()
    input.readline()
    return proj_name, mean_steps, mean_time

def read_test_result(input, test, size, projs):
    '''Read the results of all projs.
    '''
    line = '%s & %d' % (test, size)
    for i in range(len(projs)):
        proj_name, mean_steps, mean_time = read_projection_result(input)
        line += ' & %.2f & %.3f' % (mean_steps, 1000.0*mean_time)
    line += r'\\ \hline' + '\n'
    return line

def read_tables(input):
    '''Build contents of table 1 (with Newton, Secant, Var. Fixing and
    Median Search).
    '''
    test_order = (('UCI Adult', 1065),
                  ('UCI Adult', 2265),
                  ('UCI Adult', 3185),
                  ('UCI Adult', 6370),
                  ('UCI Adult', 12740),
                  ('MNIST', 800),
                  ('MNIST', 1600),
                  ('MNIST', 3200),
                  ('MNIST', 6400),
                  ('MNIST', 11702))
        
    projs1 = ['Newton', 'Newton NC', 'Secant', 'BLG', 'Var. Fixing', 'Median Search']
    projs2 = ['Newton Warm', 'Newton Warm NC', 'Secant Warm', 'BLG Warm']

    table1, table2 = '', ''
    for test, size in test_order:
        input.readline()
        table1 += read_test_result(input, test, size, projs1)
        table2 += read_test_result(input, test, size, projs2)
    return table1, table2

def main():
    '''Build tables 1 and 2.
    '''
    try:
        input = file(sys.argv[1])
    except:
        print 'Usage summarize <log file>'
        sys.exit(1)

    table1, table2 = read_tables(input)
    print LATEXSTART
    print HEADTABLE1
    print table1
    print BUTTONTABLE1
    print
    print HEADTABLE2
    print table2
    print BUTTONTABLE2
    print LATEXEND
    
if __name__ == '__main__':
    main()

