/*
 * Module Name: svmprob
 *
 * Description: Main program to use SPG code to solve the SVM training
 *    problem described in SECTION 6.2 of the manuscript
 *
 *    R. Cominetti, W. F. Mascarenhas, and Paulo J. S. Silva. "A
 *    Newton method for the continuous quadratic knapsack problem".
 *
 * Copyright: Paulo J. S. Silva <pjssilva@gmail.com> 2012.
 */

#include <stdlib.h>
#include <stdio.h>

#include "../../third_party/spg/spg.h"
#include "svmprob.h"

int main (int argc, char *argv[]) {

    /* Get options from command line */
    if (argc < 5) {
        fprintf(stderr, "Usage: svmma <problem> <projection> <repetitions> <C>\n");
        fprintf(stderr, "Possible values for projection: fix, search, newton, newtonwarm, secantwarm\n");
        exit(1);
    }

    /* Read and initialize problem data */
    int n; /* Problem dimension         */
    FILE *problem = fopen(argv[1], "r");
    printf("Reading problem data...\n");
    if (problem == NULL) {
        fprintf(stderr, "Problem %s does not exist... exiting.\n", argv[1]);
        exit(1);
    }
    FSCANF(problem, "%4c", 1, (char *)(&n));
    FSCANF(problem, "%4c", 1, (char *)(&n));
    fclose(problem);
    read_problem(argv[1], atof(argv[4]), argv[2], (unsigned) atoi(argv[3]));
    
    /* Get memory */
    /* Current point (multiplier) */
    double *restrict x = (double *) malloc (n * sizeof(double));
    /* A gradient */
    double *restrict g = (double *) malloc (n * sizeof(double));
    
    /* Define initial Guess */
    for (unsigned i = 0; i < n; ++i)
        x[i] = -1.0;
    
    /* Set solver parameters */
    int 
        iprint = 0,          /* SPG should not print to stdout or file */
        maxit  = 10000000,   /* Maximum number of SPG iterations allowed */ 
        maxfc  = 10 * maxit; /* Maximum number of function evaluations */
    double
        epsopt = 1.0e-5;     /* Precision required from projected gradient */

    /* Solver outputs */
    int
        fcnt,    /* Number of functions calls used by SPG. */
        inform,  /* Indicate error in evalf, evalg or proj. */
        spginfo, /* Reason why SPG stopped */
        iter;    /* Number of SPG iterations */
    double
        f,       /* Function value at solution x */
        gpsupn;  /* Inifinty norm of the proejected gradient at x */
    /* Call SPG */
    printf("Calling SPG...\n");
    spg(n,x,epsopt,maxit,maxfc,iprint,&f,&gpsupn,&iter,&fcnt,&spginfo,&inform);
    
    /* Save solution */
    FILE *fp = fopen ("solution.txt","w");
    for (unsigned i = 0; i < n; ++i)
        fprintf(fp,"x[%d] = %e\n",i, x[i]);
    fclose(fp);
    
    /* Save statistics */
    printStatistics();
    printf("SPG Statistics: %d %d %d %e %e %d\n",n,iter,fcnt,f,gpsupn,spginfo);

    /* Free memory */
    free(x);
    free(g);
    free_problem();
    
    return 0;
}


